import sys
import os
import linecache
from importlib.abc import MetaPathFinder, Loader
from importlib.util import spec_from_loader

from koatl import transpile


class TlFinder(MetaPathFinder):
    def find_spec(self, fullname, path, target=None):
        if path is None:
            path = sys.path

        module_name = fullname.split(".")[-1]

        for entry in path:
            # 1. Check for a regular module file: {entry}/{module_name}.tl
            file_path = os.path.join(entry, f"{module_name}.tl")
            if os.path.isfile(file_path):
                return spec_from_loader(fullname, TlLoader(file_path))

            # 2. Check for a package: {entry}/{module_name}/__init__.tl
            package_path = os.path.join(entry, module_name)
            init_path = os.path.join(package_path, "__init__.tl")

            if os.path.isdir(package_path) and os.path.isfile(init_path):
                spec = spec_from_loader(fullname, TlLoader(init_path), is_package=True)
                spec.submodule_search_locations = [package_path]
                return spec

        return None


class TlLoader(Loader):
    def __init__(self, filepath):
        self.filepath = filepath

    def create_module(self, spec):
        return None

    def exec_module(self, module):
        module.__file__ = self.filepath

        with open(self.filepath, "r") as f:
            source_code = f.read()

        linecache.cache[self.filepath] = (
            len(source_code),
            None,
            [line + "\n" for line in source_code.splitlines()],
            self.filepath,
        )

        if module.__name__.startswith("koatl.prelude"):
            transpiled_code = transpile(
                source_code, mode="no_prelude", filename=self.filepath
            )
        else:
            transpiled_code = transpile(
                source_code, mode="module", filename=self.filepath
            )

        code = compile(transpiled_code, self.filepath, "exec")

        exec(code, module.__dict__)


def install_hook():
    sys.meta_path.insert(0, TlFinder())



from functools import partial
from itertools import count
from types import new_class
from .._rs import fast_vget, fast_vset, fast_vset_trait


def ExtensionProperty(f):
    f.ext_prop = True
    return f


def vget(obj, name):
    try:
        return getattr(obj, name)
    except:
        # special case for iter - this could be implemented using types and trait vtbls
        # but this is simpler and probably faster
        if name == "iter":
            if isinstance(obj, slice):
                start = obj.start if obj.start is not None else 0
                step = obj.step if obj.step is not None else 1
                if obj.stop is None:
                    return count(start, step)
                else:
                    return range(start, obj.stop, step)

            try:
                return obj.items()
            except AttributeError:
                pass

            try:
                return iter(obj)
            except TypeError:
                pass

        v = fast_vget(obj, name)
        if v is not None:
            if hasattr(v, "ext_prop"):
                return v(obj)

            return partial(v, obj)

        raise AttributeError(
            f"'{type(obj).__name__}' object has no v-attribute '{name}'"
        ) from None


def vhas(obj, name):
    if hasattr(obj, name):
        return True

    if name == "iter":
        if isinstance(obj, slice):
            return True

        try:
            obj.items()
            return True
        except AttributeError:
            pass

        try:
            iter(obj)
            return True
        except TypeError:
            pass

    v = fast_vget(obj, name)
    if v is not None:
        return True

    return False


def Trait(module, name, methods, *, requires=[]):
    def fix_methods(type_name, methods):
        import inspect

        for method_name, method in methods.items():
            method.__name__ = method_name
            method.__qualname__ = f"{type_name}.{method_name}"

            if not isinstance(method, (staticmethod, classmethod)):
                sig = inspect.signature(method)
                params = list(sig.parameters.values())
                params[0].replace(name="self")
                sig = sig.replace(parameters=params)
                method.__signature__ = sig

    fix_methods(name, methods)

    def instancecheck(cls, instance):
        if cls != typ:
            # if not checking for the trait itself, use the default behavior
            return type.__instancecheck__(cls, instance)

        for req in requires:
            try:
                vget(instance, req)
            except AttributeError:
                return False

        return True

    meta = type(
        f"{name}Meta",
        (type,),
        {"__instancecheck__": instancecheck, "__module__": "types"},
    )

    def populate(ns):
        for method_name, method in methods.items():
            ns[method_name] = method
        ns["__module__"] = module
        ns["_methods"] = methods
        ns["_requires"] = requires

    typ = new_class(name, (), {"metaclass": meta}, populate)

    return typ


def register_global_attr(type, name, value):
    fast_vset(type, name, value)


def register_global_trait(type):
    for name, method in type._methods.items():
        fast_vset_trait(type.__name__, type._requires, name, method)


__all__ = [
    "Trait",
    "ExtensionProperty",
    "register_global_attr",
    "register_global_trait",
]



from functools import wraps
from .virtual import vget
from .record import Record


__all__ = ["MatchError"]


class MatchError(Exception):
    def __init__(self, message):
        super().__init__(message)


def set_exports(package_name, globals_dict, exports, module_star_exports):
    import importlib

    exports = set(exports)

    for module in module_star_exports:
        mod = importlib.import_module(module, package_name)

        if hasattr(mod, "__all__"):
            for name in mod.__all__:
                exports.add(name)
        else:
            for name in dir(mod):
                if name.startswith("_"):
                    continue

                exports.add(name)

    if "__all__" not in globals_dict:
        globals_dict["__all__"] = ()

    globals_dict["__all__"] = tuple(set(globals_dict["__all__"]) | exports)


def unpack_record(obj):
    """
    used in record unpacking
    """
    if hasattr(obj, "items"):
        return Record(obj.items())
    else:
        return Record(obj.__dict__)


def ok(obj):
    if obj is None:
        return False
    if isinstance(obj, BaseException):
        return False
    return True


def do(f):
    @wraps(f)
    def impl(*args, **kwargs):
        gen = f(*args, **kwargs)

        try:
            m = gen.send(None)
        except StopIteration as e:
            return e.value

        def recurse(v):
            nonlocal m
            try:
                m = gen.send(v)
                return vget(m, "bind_once")(recurse)
            except StopIteration as e:
                try:
                    return vget(m, "pure")(e.value)
                except AttributeError:
                    return e.value

        try:
            # TODO: this is a workaround to avoid recursion.
            # is it possible to get bind_gen directly from bind_once?

            return vget(m, "bind_gen")(gen)
        except (NotImplementedError, AttributeError):
            return vget(m, "bind_once")(recurse)

    return impl



from IPython.core.inputtransformer2 import (
    EscapedCommand,
    MagicAssign,
    SystemAssign,
    find_end_of_continued_line,
    assemble_continued_line,
)

import re

ESC_SHELL = "!"  # Send line to underlying system shell
ESC_SH_CAP = "!!"  # Send line to system shell and capture output
ESC_HELP = "?"  # Find information about object
ESC_HELP2 = "??"  # Find extra-detailed information about object
ESC_MAGIC = "%"  # Call magic function
ESC_MAGIC2 = "%%"  # Call cell-magic function
ESC_QUOTE = ","  # Split args on whitespace, quote each as string and call
ESC_QUOTE2 = ";"  # Quote all args as a single string, call
ESC_PAREN = "/"  # Call first argument with rest of line as arguments

ESCAPE_SINGLES = {"!", "?", "%", ",", ";", "/"}
ESCAPE_DOUBLES = {"!!", "??"}  # %% (cell magic) is handled separately


def _make_help_call(target, esc):
    """Prepares a pinfo(2)/psearch call from a target name and the escape
    (i.e. ? or ??)"""
    method = "pinfo2" if esc == "??" else "psearch" if "*" in target else "pinfo"
    arg = " ".join([method, target])
    # Prepare arguments for get_ipython().run_line_magic(magic_name, magic_args)
    t_magic_name, _, t_magic_arg_s = arg.partition(" ")
    t_magic_name = t_magic_name.lstrip(ESC_MAGIC)
    return "get_ipython().run_line_magic({}, {})".format(
        crepr(t_magic_name), crepr(t_magic_arg_s)
    )


def crepr(s):
    assert isinstance(s, str)
    return '"' + repr(s)[1:-1] + '"'


def _tr_help(content):
    """Translate lines escaped with: ?

    A naked help line should fire the intro help screen (shell.show_usage())
    """
    if not content:
        return "get_ipython().show_usage()"

    return _make_help_call(content, "?")


def _tr_help2(content):
    """Translate lines escaped with: ??

    A naked help line should fire the intro help screen (shell.show_usage())
    """
    if not content:
        return "get_ipython().show_usage()"

    return _make_help_call(content, "??")


def _tr_magic(content):
    "Translate lines escaped with a percent sign: %"
    name, _, args = content.partition(" ")
    return "get_ipython().run_line_magic({}, {})".format(crepr(name), crepr(args))


def _tr_quote(content):
    "Translate lines escaped with a comma: ,"
    name, _, args = content.partition(" ")
    return '%s("%s")' % (name, '", "'.join(args.split()))


def _tr_quote2(content):
    "Translate lines escaped with a semicolon: ;"
    name, _, args = content.partition(" ")
    return '%s("%s")' % (name, args)


def _tr_paren(content):
    "Translate lines escaped with a slash: /"
    name, _, args = content.partition(" ")
    if name == "":
        raise SyntaxError(f'"{ESC_SHELL}" must be followed by a callable name')

    return "%s(%s)" % (name, ", ".join(args.split()))


tr = {
    ESC_SHELL: lambda s: "get_ipython().system({})".format(crepr(s)),
    ESC_SH_CAP: lambda s: "get_ipython().getoutput({})".format(crepr(s)),
    ESC_HELP: _tr_help,
    ESC_HELP2: _tr_help2,
    ESC_MAGIC: _tr_magic,
    ESC_QUOTE: _tr_quote,
    ESC_QUOTE2: _tr_quote2,
    ESC_PAREN: _tr_paren,
}


class KoatlEscapedCommand(EscapedCommand):
    def transform(self, lines):
        """Transform an escaped line found by the ``find()`` classmethod."""
        start_line, start_col = self.start_line, self.start_col

        indent = lines[start_line][:start_col]
        end_line = find_end_of_continued_line(lines, start_line)
        line = assemble_continued_line(lines, (start_line, start_col), end_line)

        if len(line) > 1 and line[:2] in ESCAPE_DOUBLES:
            escape, content = line[:2], line[2:]
        else:
            escape, content = line[:1], line[1:]

        if escape in tr:
            call = tr[escape](content)
        else:
            call = ""

        lines_before = lines[:start_line]
        new_line = indent + call + "\n"
        lines_after = lines[end_line + 1 :]

        return lines_before + [new_line] + lines_after


class KoatlMagicAssign(MagicAssign):
    priority = 9

    def transform(self, lines):
        start_line, start_col = self.start_line, self.start_col
        lhs = lines[start_line][:start_col]
        end_line = find_end_of_continued_line(lines, start_line)
        rhs = assemble_continued_line(lines, (start_line, start_col), end_line)
        assert rhs.startswith("%"), rhs
        magic_name, _, args = rhs[1:].partition(" ")

        lines_before = lines[:start_line]
        call = "get_ipython().run_line_magic({}, {})".format(
            crepr(magic_name), crepr(args)
        )
        new_line = lhs + call + "\n"
        lines_after = lines[end_line + 1 :]

        return lines_before + [new_line] + lines_after


class KoatlSystemAssign(SystemAssign):
    priority = 9

    def transform(self, lines):
        start_line, start_col = self.start_line, self.start_col

        lhs = lines[start_line][:start_col]
        end_line = find_end_of_continued_line(lines, start_line)
        rhs = assemble_continued_line(lines, (start_line, start_col), end_line)
        assert rhs.startswith("!"), rhs
        cmd = rhs[1:]

        lines_before = lines[:start_line]
        call = "get_ipython().getoutput({})".format(crepr(cmd))
        new_line = lhs + call + "\n"
        lines_after = lines[end_line + 1 :]

        return lines_before + [new_line] + lines_after


def koatl_cell_magic(lines):
    if not lines or not lines[0].startswith("%%"):
        return lines
    if re.match(r"%%\w+\?", lines[0]):
        # This case will be handled by help_end
        return lines
    magic_name, _, first_line = lines[0][2:].rstrip().partition(" ")
    body = "".join(lines[1:])
    return [
        "get_ipython().run_cell_magic({}, {}, {})\n".format(
            crepr(magic_name), crepr(first_line), crepr(body)
        )
    ]



import pytest
import sys
import koatl.cli
from pathlib import Path

sys.path.append(str(Path(__file__).parent / "e2e"))


def get_test_data(dirs):
    data_dirs = [
        Path(__file__).parent / "e2e" / dirs,
    ]

    test_cases = []
    for data_dir in data_dirs:
        for file_path in data_dir.glob("*.tl"):
            test_cases.append(pytest.param(file_path, id=str(file_path)))

    return test_cases


@pytest.mark.parametrize("test_file", get_test_data("base"))
def test_e2e_native_emit_base(test_file):
    e2e_native_emit(test_file, "no_prelude")


@pytest.mark.parametrize("test_file", get_test_data("prelude"))
def test_e2e_native_emit_prelude(test_file):
    e2e_native_emit(test_file, "script")


@pytest.mark.parametrize("test_file", get_test_data("base"))
def test_e2e_base(test_file):
    e2e(test_file, "no_prelude")


@pytest.mark.parametrize("test_file", get_test_data("prelude"))
def test_e2e_prelude(test_file):
    e2e(test_file, "script")


def e2e_native_emit(test_file, mode):
    import linecache

    with open(test_file, "r") as f:
        source = f.read()
    source, source_map = koatl.transpile_raw(source, mode=mode)

    global_dict = {}

    try:
        linecache.cache["<string>"] = (
            len(source),
            None,
            source.splitlines(),
            "<string>",
        )
        codeobj = compile(source, "<string>", "exec")
        exec(codeobj, global_dict, global_dict)
    except Exception as e:
        print(source)
        raise

    print("end", test_file)


def e2e(test_file, mode):
    koatl.cli.run_from_path(test_file, mode=mode)



"""
This module initializes the environment for koatl.
It sets up the meta-finder hook to enable importing .tl files,
and also declares functions that are required for certain tl features to work,
such as coalescing and module exports.

koatl.runtime should be written in Python only since otherwise
it would create a circular dependency.
"""

import functools
from types import SimpleNamespace

from . import meta_finder

meta_finder.install_hook()


from .virtual import *
from .record import *
from .helpers import *


def dummy_memo(*args, **kwargs):
    raise RuntimeError(
        "memo is not available without the prelude. Please import koatl.prelude."
    )


__tl__ = SimpleNamespace(
    Exception=Exception,
    slice=slice,
    vget=virtual.vget,
    vhas=virtual.vhas,
    memo=dummy_memo,
    unpack_record=helpers.unpack_record,
    set_exports=helpers.set_exports,
    do=helpers.do,
    ok=helpers.ok,
    partial=functools.partial,
    **{name: helpers.__dict__[name] for name in helpers.__all__},
    **{name: record.__dict__[name] for name in record.__all__},
    **{name: virtual.__dict__[name] for name in virtual.__all__}
)


__all__ = [
    "__tl__",
    *helpers.__all__,
    *record.__all__,
    *virtual.__all__,
]
